var FractalPageObject = (function (exports) {
  'use strict';

  let root;
  function getEmberTestRoot() {
      try {
          return window.require('@ember/test-helpers').getRootElement();
      }
      catch (e) {
          return null;
      }
  }
  /**
   * Set the global default root element -- by default {@link PageObject}s will
   * only match elements that are descendants of this element. The default root
   * element is `document.body`.
   *
   * @param {Element|Function} element the root element or a function that will
   * return it
   */
  function setRoot(element) {
      root = element;
  }
  /**
   * Get the global default root element
   *
   * @returns the root element
   *
   * @private
   */
  function getRoot() {
      if (root instanceof Element) {
          return root;
      }
      else {
          return getEmberTestRoot() || document.body;
      }
  }

  /* istanbul ignore file */
  /**
   * Base class for {@link PageObject} that contains stub implementations of a
   * subset of array methods (the ones that produce a possibly useful value). The
   * actual functionality is handled by the proxy, the implementations aren't
   * important, these are just here for type checking.
   *
   * It really seems like there should be a better way to do this, but I can't
   * figure out how else to make PageObject subclasses look like arrays of
   * `WithElement<this>`.
   *
   * @private
   */
  class ArrayStub {
      constructor() {
          //
          // Array API
          //
          // Here we provide typed stub implementations for a subset of array methods
          // (the ones that produce a possibly useful value). The actual functionality
          // is handled by the proxy, the implementations aren't important, these are
          // just here for type checking.
          //
          // It really seems like there should be a better way to do this, but I can't
          // figure out how else to make PageObject subclasses look like arrays of
          // `WithElement<this>`.
          //
          /**
           * @private
           */
          this.length = 0;
      }
      /**
       * @private
       */
      pop() {
          return undefined;
      }
      /**
       * @private
       */
      concat(..._items) {
          return [];
      }
      /**
       * @private
       */
      reverse() {
          return [];
      }
      /**
       * @private
       */
      shift() {
          return undefined;
      }
      /**
       * @private
       */
      slice(_start, _end) {
          return [];
      }
      /**
       * @private
       */
      sort(_compareFn) {
          return [];
      }
      /**
       * @private
       */
      indexOf(_searchElement, _fromIndex) {
          return -1;
      }
      /**
       * @private
       */
      lastIndexOf(_searchElement, _fromIndex) {
          return -1;
      }
      /**
       * @private
       */
      every(_predicate, _thisArg) {
          return false;
      }
      /**
       * @private
       */
      some(_predicate, _thisArg) {
          return false;
      }
      /**
       * @private
       */
      forEach(_callbackfn, _thisArg) { }
      /**
       * @private
       */
      map(_callbackfn, _thisArg) {
          return [];
      }
      /**
       * @private
       */
      filter(_predicate, _thisArg) {
          return [];
      }
      /**
       * @private
       */
      reduce(_callbackfn, initialValue) {
          return initialValue;
      }
      /**
       * @private
       */
      reduceRight(_callbackfn, initialValue) {
          return initialValue;
      }
      /**
       * @private
       */
      find(_predicate, _thisArg) {
          return undefined;
      }
      /**
       * @private
       */
      findIndex(_predicate, _thisArg) {
          return -1;
      }
      /**
       * @private
       */
      includes(_searchElement, _fromIndex) {
          return false;
      }
      /**
       * @private
       */
      [Symbol.iterator]() {
          return {};
      }
      /**
       * @private
       */
      entries() {
          return {};
      }
      /**
       * @private
       */
      keys() {
          return {};
      }
      /**
       * @private
       */
      values() {
          return {};
      }
  }

  /**
   * A class that can execute selector-based DOM queries, with supporting for
   * intermediate indexing.
   */
  class DOMQuery {
      /**
       * @param root the root element from which to query
       * @param selector the selector to query
       */
      constructor(root, selector) {
          this.root = root;
          this.selector = selector;
      }
      /**
       * Query the first matching element
       */
      query() {
          if (!this.root) {
              return null;
          }
          if (this.selector) {
              return this.root.querySelector(this.selector);
          }
          else {
              return this.root;
          }
      }
      /**
       * Query all matching elements
       */
      queryAll() {
          if (!this.root) {
              return [];
          }
          if (this.selector) {
              return Array.from(this.root.querySelectorAll(this.selector));
          }
          else {
              return [this.root];
          }
      }
      /**
       * Create a child {@link DOMQuery} from a selector and an optional index.
       * {@link DOMQuery} objects created using this method are not guaranteed to be
       * lazy -- if an index is specified, the query will be executed immediately.
       *
       * @param selector selector to query
       * @param index target index of the {@link Element} in the query results, or
       * null if this isn't an index query
       */
      createChild(selector, index) {
          // Create a DOMQuery just using the selector, and ignoring the index, which
          // is accomplished by appending the new selector to ours and keeping the
          // same root.
          let child = new DOMQuery(this.root, `${this.selector} ${selector}`.trim());
          if (index !== null) {
              // We have an index, so we have to query all matching elements to find the
              // one we are targeting, and set that as our root with no selector.
              let element = child.queryAll()[index] || null;
              return new DOMQuery(element, '');
          }
          else {
              // No index, so nothing extra to do
              return child;
          }
      }
  }

  /**
   * A factory for creating {@link PageObject}s. The factory is constructed with a
   * selector and an optional class that must be a {@link PageObject} subclass.
   * The {@link PageObjectFactor#create} method instantiates page objects using
   * the data provided to the factoryt.
   */
  class PageObjectFactory {
      /**
       * @param selector the selector for page objects created from this factory
       * @param Class the class to use when creating page objects. It must be a
       * subclass of {@link PageObject}. If not supplied, {@link PageObject} will be
       * used.
       */
      constructor(selector, Class) {
          this.selector = selector;
          this.Class = Class;
      }
      /**
       * Create a {@link PageObject}
       *
       * @param parent the {@link PageObject} to set as the new page object's parent
       * @returns the new page object
       */
      create(parent) {
          let Class = this.Class || PageObject;
          return new Class(this.selector, parent);
      }
  }

  const DOM_QUERY = Symbol('DOM query');
  const CLONE_WITH_INDEX = Symbol('withIndex');

  /**
   * Create a proxy wrapping a {@link PageObject} and implementing array
   * functionality.
   *
   * @param PageObject the page object to wrap in a proxy
   *
   * @returns the proxy implementing the page object & array functionality
   */
  function createProxy(pageObject) {
      return new Proxy(pageObject, {
          get(pageObject, prop, receiver) {
              let index = convertToInt(prop);
              if (index !== null) {
                  // `page.foo[1]` -- clone the page object with an index. We special case
                  // this rather than handling it like the rest of the Array API because
                  // we want users to be use an index for which there isn't actually a
                  // matching element in the DOM and still get a valid page object (just
                  // without an element), e.g. `assert.notOk(page.foo[3].element)`
                  return pageObject[CLONE_WITH_INDEX](index);
              }
              else if (Reflect.has([], prop) && prop !== 'constructor') {
                  // Array API method/property, so create an array with page objects with
                  // indices wrapping all of our matching elements and apply the
                  // method/property to it. It should be possible to populate PageObject
                  // (since it's an Array subclass) with the child page objects and
                  // operate on that, but if we do this, something about how we set up
                  // PageObjects and their prototypes, and possibly how the code in
                  // transpiled, causes `pageObject.map(...)` to construct and return a
                  // PageObject with its parent set to an integer (instead of a page
                  // object parent).
                  let elements = pageObject.elements;
                  let children = elements.map((_e, i) => pageObject[CLONE_WITH_INDEX](i));
                  let value = children[prop];
                  if (typeof value === 'function') {
                      return (...args) => value.call(children, ...args);
                  }
                  else {
                      return value;
                  }
              }
              else {
                  // Not the Array API, so either something in our base PageObject
                  // implementation, e.g. `element` or `elements`, or a property/method on
                  // whatever subclass of PageObject `pageObject` is. So, get the value
                  // with factory support
                  return getWithFactorySupport(pageObject, prop, receiver);
              }
          },
      });
  }
  /**
   * A child-node-aware wrapper for Reflect.get() that detects when the return
   * value is a {@link PageObjectFactory} and uses it to instantiate a
   * {@link PageObject} under the given parent
   *
   * @param pageObject the page object whose property we're getting
   * @param prop the property name to get
   * @param receiver the proxy receiver
   *
   * @returns the resolved property value, which will be a constructed
   * {@link PageObject} if the actual property value is a
   * {@link PageObjectFactory}
   */
  function getWithFactorySupport(pageObject, prop, receiver) {
      let value = Reflect.get(pageObject, prop, receiver);
      if (value instanceof PageObjectFactory) {
          // PageObjetFactory, so use it to instantiate a PageObject that is a child
          // of this PageObject
          return value.create(pageObject);
      }
      else {
          // Some other value -- just return it
          return value;
      }
  }
  // Thanks @pzuraq (copied from tracked-built-ins)
  function convertToInt(prop) {
      if (typeof prop === 'symbol')
          return null;
      let num = Number(prop);
      if (isNaN(num))
          return null;
      return num % 1 === 0 ? num : null;
  }

  /**
   * This class implements all the basic page object functionality, and all page
   * objects must inherit from it. It can host {@link selector} and
   * {@link globalSelector} fields, and will properly instantiate them as nested
   * {@link PageObject}s when accessed. Each page object represents a DOM query
   * that matches zero or more {@link Element}s.
   *
   * {@link PageObject}s exist in a tree where each {@link PageObject}'s elements
   * are descendants of its parent's elements. The root of the tree is a top-level
   * {@link PageObject} created using `new PageObject()`, that by default will
   * match the root element (the body element or whatever was set as the root
   * using {@link setRoot}). It can also be constructed with a selector argument,
   * `new PageObject(selector)`, in which case its query will be a selector query
   * starting from the root element. Then each non-root {@link PageObject} has a
   * selector and optional index that define its elements relative to its parent's
   * elements.
   *
   * As long as no page objects have an index, each {@link PageObject}
   * conceptually appends its selector to its parent's selector, in effect making
   * its matching elements
   *
   * ```javascript
   * rootElement.querySelectorAll(`${this.parent.selector} ${this.selector}`);
   * ```
   *
   * Page objects that do have an index ("indexed page objects") restrict their
   * matching elements to only the one element at their index within their full
   * query results (or no elements if there is no element at that index). That in
   * effect "resets" the root element for its descendants, so its children execute
   * queries relative to its matching element (if it has one) and so on.
   *
   * {@link PageObject}s are lazy, meaning that their query is not evaluated when
   * they are constructed, but is evaluated and re-evaluated each time a property
   * that depends on it is accessed.
   *
   * {@link PageObject}s expose an API for interacting with their matching
   * elements that comprises {@link PageObject#element},
   * {@link PageObject#elements}, and an {@link Array} API that exposes the page
   * object's matching {@link Element}s wrapped in indexed {@link PageObject}s.
   * The index operator will return an indexed {@link PageObject} that may or may
   * not match an element (similar to how you can index off the end of a native
   * array and get `undefined`), while various array iteration methods like
   * {@link PageObject#map} generate a range of {@link PageObject}s that reflect
   * only the indices that actually match an element.
   *
   * Descendant {@link PageObject}s are defined by subclassing {@link PageObject}
   * and using the {@link selector} factory function to initialize class fields.
   *
   * When creating a top-level {@link PageObject} directly using `new`, its query
   * will match the root element (the body element or whatever was )
   *
   * @param {string} [selector] the selector to use for this page object's query
   * @param parent (for internal use only)
   * @param index (for internal use only)
   * @param rootElement (for internal use only)
   *
   * @example
   *
   * import { PageObject, selector, setRoot } from 'fractal-page-object';
   *
   * class Page extends PageObject {
   *   list = selector('.list');
   * }
   *
   * setRoot(rootElement);
   *
   * let page = new Page();
   * page.list.elements; // rootElement.querySelectorAll('.list')
   *
   * @example
   *
   * import { PageObject, selector, setRoot } from 'fractal-page-object';
   *
   * class Page extends PageObject {
   *   list = selector('.list');
   * }
   *
   * setRoot(rootElement);
   *
   * let page = new Page('.container');
   * page.list.elements; // rootElement.querySelectorAll('.container .list')
   */
  class PageObject extends ArrayStub {
      constructor(selector = '', parent = null, index = null, rootElement = null) {
          super();
          this.selector = selector;
          this.parent = parent;
          this.index = index;
          this.rootElement = rootElement;
          return createProxy(this);
      }
      /**
       * This page object's single matching DOM element -- the first DOM element
       * matching this page object's query if this page object does not have an
       * index, or the `index`th matching DOM element if it does have an index
       * specified.
       *
       * @type {Element | null}
       */
      get element() {
          return this[DOM_QUERY].query();
      }
      /**
       * This page object's list of matching DOM elements. If this page object has
       * an index, this property will always have a length of 0 or 1.
       *
       * @type {Element[]}
       */
      get elements() {
          return this[DOM_QUERY].queryAll();
      }
      /**
       * A {@link DOMQuery} object for querying this page object's matching DOM
       * elements
       *
       * @private
       */
      get [DOM_QUERY]() {
          let parentQuery;
          if (this.parent instanceof PageObject) {
              parentQuery = this.parent[DOM_QUERY];
          }
          else if (this.parent instanceof Element) {
              parentQuery = new DOMQuery(this.parent, '');
          }
          else {
              parentQuery = new DOMQuery(getRoot(), '');
          }
          return parentQuery.createChild(this.selector, this.index);
      }
      /**
       * Create a clone of this {@link PageObject}, but with an index specified. If
       * called on an instance of a subclass of {@link PageObject}, the clone will
       * be an instance of that subclass.
       *
       * @param index the index
       *
       * @returns a clone of this page object, but only selecting the DOM element in
       * the matching elements at the given index
       *
       * @private
       */
      [CLONE_WITH_INDEX](index) {
          // We have to be careful about how we treat the index. If this page object
          // already has an index, then it only matches 0 elements or 1 element, so
          // cloning it with a new index can only possibly match an element if that
          // new index is 0. For example, suppose `pageObject` matched two DOM
          // elements. Then `pageObject[0]` would match only the first DOM element,
          // and `pageObject[0][0]` would match that same DOM element, while
          // `pageObject[0][1]` wouldn't match any DOM element. So we need to factor
          // in this page object's index when determining what index in the full query
          // result set the passed-in index is actually referring to.
          let resolvedIndex;
          if (this.index === null) {
              // No index, so the index stays as provided
              resolvedIndex = index;
          }
          else {
              // We have an index. So if the provided index is 0, the clone should match
              // the same element as this page object, and should get the same index. If
              // the provided index is anything else, it can never match any DOM elements,
              // so we'll give the clone an index of -1 so it never matches.
              if (index === 0) {
                  resolvedIndex = this.index;
              }
              else {
                  resolvedIndex = -1;
              }
          }
          let Class = this.constructor;
          return new Class(this.selector, this.parent, resolvedIndex, this.rootElement);
      }
  }

  function isPageObjectSubclass(Class) {
      return Class === PageObject || Class.prototype instanceof PageObject;
  }
  function validateSelectorArguments(selector, Class) {
      if (!selector) {
          throw new Error('Cannot specify an empty selector');
      }
      if (Class && !isPageObjectSubclass(Class)) {
          throw new Error('Custom globalSelector() class must be PageObject or PageObject subclass');
      }
  }

  /**
   * Define a child PageObject. It can optionally be supplied with a
   * {@link PageObject} subclass definition to allow customizing functionality on
   * the page object, such as defining (grand)child page objects, or other helpful
   * properties and functions.
   *
   * @param {string} selector the selector relative to the parent node
   * @param {Function<PageObject>} [Class] optional {@link PageObject} subclass that
   * can be used to extend the functionality of this page object
   *
   * @returns {PageObject} a {@link PageObject} or {@link PageObject} subclass
   * instance
   *
   * @example
   *
   * import { PageObject, selector } from 'fractal-page-object';
   *
   * class Page extends PageObject {
   *   list = selector('.list', class extends PageObject {
   *     items = selector('li');
   *   });
   * }
   * let page = new Page();
   * page.list.elements; // document.body.querySelectorAll('.list')
   * page.list.items.elements; // document.body.querySelectorAll('.list li')
   */
  function selector(selector, Class) {
      validateSelectorArguments(selector, Class);
      // Return a factory, but typed as the class it will instantiate since the
      // proxy will do the intantiation anytime a property containing a factory is
      // accessed
      return new PageObjectFactory(selector, Class);
  }

  /**
   * A factory for creating {@link PageObject}s in a global scope, i.e. whose queries
   * are not scoped to their parent. The factory has an optional rootElement argument
   * that specifies the element under which to scope queries. If not specified, it will
   * default to the global root set via {@link setRoot}.
   */
  class GlobalPageObjectFactory extends PageObjectFactory {
      /**
       * @param selector the selector for page objects created from this factory
       * @param rootElement the element that will be used to scope the page object's
       * selector query, defaulting to the global root if not specified
       * @param Class the class to use when creating page objects. It must be a
       * subclass of {@link PageObject}. If not supplied, {@link PageObject} will be
       * used.
       */
      constructor(selector, rootElement, Class) {
          super(selector, Class);
          this.rootElement = rootElement;
      }
      /**
       * Create a {@link PageObject}
       *
       * @param parent the {@link PageObject} to set as the new page object's parent
       * @returns the new page object
       */
      create() {
          return super.create(this.rootElement);
      }
  }

  /**
   * Define a {@link PageObject} with a global scope, i.e. not scoped by its
   * parent page object. Useful for cases like popovers and dropdowns, where the
   * UI control is logically inside a given component, but all or part of it
   * renders elsewhere in the DOM, such as directly under the body.
   * {@link globalSelector} accepts a selector and optional custom class like
   * {@link selector()}, but the queries of the page objects it generates will be
   * executed from the root (`document.body` or whatever was passed to
   * {@link setRoot}) rather than the parent page object's elements.
   *
   * In some cases, it makes sense to set up a global root using {@link setRoot},
   * but some elements might be rendered outside even that, such as directly under
   * the body. For cases like this, {@link globalSelector} accepts an optional
   * root element to use as the root of the queries of the page object it
   * generates.
   *
   * @param {string} selector the selector
   * @param {Function<PageObject>} Class optional {@link PageObject} subclass that
   * can be used to extend the functionality of this page object
   *
   * @returns {PageObject} a {@link PageObject} or {@link PageObject} subclass
   * instance
   *
   * @example
   *
   * import { PageObject, selector, globalSelector } from 'fractal-page-object';
   *
   * const testContainer = document.querySelector('#test-container');
   * setRoot(testContainer);
   *
   * class Page extends PageObject {
   *   listItems = selector('.list-item', class extends PageObject {
   *     popover = globalSelector('.popover', class extends PageObject {
   *       icon = selector('.icon');
   *     });
   *   });
   * }
   * let page = new Page();
   * page.listItems[0]; // testContainer.querySelectorAll('.listItems')[0]
   * page.listItems[0].popover; // testContainer.querySelectorAll('.popover')
   * page.listItems[0].popover.icon; // testContainer.querySelectorAll('.popover .icon')
   *
   * @example
   *
   * import { setRoot, PageObject, selector, globalSelector } from 'fractal-page-object';
   *
   * const testContainer = document.querySelector('#test-container');
   * setRoot(testContainer);
   *
   * class Page extends PageObject {
   *   listItems = selector('.list-item', class extends PageObject {
   *     popover = globalSelector('.popover', document.body, class extends PageObject {
   *       icon = selector('.icon');
   *     });
   *   });
   * }
   * let page = new Page();
   * page.listItems[0]; // testContainer.querySelectorAll('.listItems')[0]
   * page.listItems[0].popover; // document.body.querySelectorAll('.popover')
   * page.listItems[0].popover.icon; // document.body.querySelectorAll('.popover .icon')
   */ function globalSelector(...args) {
      let selector = args[0];
      let rootElement;
      let Class;
      if (args[1] instanceof Element) {
          rootElement = args[1];
          Class = args[2];
      }
      else {
          Class = args[1];
      }
      validateSelectorArguments(selector, Class);
      // Return a factory, but typed as the class it will instantiate since the
      // proxy will do the intantiation anytime a property containing a factory is
      // accessed
      return new GlobalPageObjectFactory(selector, rootElement, Class);
  }

  exports.PageObject = PageObject;
  exports.globalSelector = globalSelector;
  exports.selector = selector;
  exports.setRoot = setRoot;

  Object.defineProperty(exports, '__esModule', { value: true });

  return exports;

}({}));
//# sourceMappingURL=fractal-page-object.js.map
